import LiteEFG
from tqdm import tqdm
import pyspiel
import csv

def train(graph, traverse_type, convergence_type, iter, print_freq, game_env="leduc_poker", output_strategy=False):
    game = pyspiel.load_game(game_env)
    env = LiteEFG.OpenSpielEnv(game, traverse_type=traverse_type, regenerate=False)
    env.set_graph(graph)

    pbar = tqdm(total=iter)
    best_exp = 1e9

    for i in range(iter):
        graph.update_graph(env)
        env.update_strategy(graph.current_strategy(), update_best=(convergence_type == "best-iterate"))
            
        if i % print_freq == 0:
            exploitability = env.exploitability(graph.current_strategy(), convergence_type)
            pbar.set_description(f'Exploitability: {exploitability[0]:.8f}, Best: {exploitability[1]:.8f}')
            pbar.update(print_freq)

    if output_strategy:
        _, df_list = env.get_strategy(graph.current_strategy(), "avg-iterate")
        for i, df in enumerate(df_list):
            df['Infoset'] = df['Infoset'].apply(lambda x: x.replace('\n', '\\n'))
            df.to_csv("strategy_" + str(i) + ".csv", quoting=csv.QUOTE_MINIMAL, quotechar='"')

def test():
    import sys
    import os
    import runpy

    package_dir = os.path.dirname(__file__)
    python_files = [file for file in os.listdir(package_dir) if file.endswith('.py') and file != '__init__.py' and file != 'utils.py']

    original_argv = sys.argv
    for file in python_files:
        module_name = f'{file[:-3]}'
        sys.argv = [module_name, '--iter', '1000', '--print_freq', '100', '--game', 'leduc_poker(suit_isomorphism=True)']
        runpy.run_module(module_name, run_name='__main__')
    
    sys.argv = original_argv

    print("Test Success")

if __name__ == "__main__":
    test()



#######################################################
# Magnetic Mirror Descent (MMD)
# Samuel Sokota, Ryan D'Orazio, J. Zico Kolter, Nicolas Loizou, Marc Lanctot, Ioannis Mitliagkas, Noam Brown, and Christian Kroer.
# "A Unified Approach to Reinforcement Learning, Quantal Response Equilibria, and Two-Player Zero-Sum Games."
# International Conference on Learning Representations (2023).
#######################################################

import LiteEFG
from typing import Literal

class graph(LiteEFG.Graph):
    def __init__(self, eta=0.005, tau=0.05, gamma=0.0, regularizer: Literal["Euclidean", "Entropy"]="Entropy", feedback="Q", weighted=False):
        super().__init__()

        self.eta = eta
        self.tau = tau
        self.gamma = gamma
        self.regularizer = regularizer
        self.feedback = feedback

        with LiteEFG.backward(is_static=True):

            self.alpha = 1.0
            if weighted:
                self.alpha = LiteEFG.const(1, 1.0)
                self.alpha.inplace(LiteEFG.aggregate(self.alpha, "sum"))
                self.alpha.inplace((self.alpha.max() + 1) * 2)

            self.ev = LiteEFG.const(1, 0.0)
            self.coef = self.tau
            self.mu = self.subtree_size.normalize(p_norm=1.0, ignore_negative=True)
            self.eta_coef = self.eta * self.coef
            
            self.u = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)

        with LiteEFG.backward():

            if(self.feedback in ["Q", "traj-Q", "counterfactual"]):
                self.full_information()
            elif self.feedback == "Outcome":
                self.outcome_sampling()

        print("===============Graph is ready for MMD===============")
        print("eta: %f, tau: %f, gamma: %f, regularizer: %s, feedback: %s" % (self.eta, self.tau, self.gamma, self.regularizer, self.feedback))
        print("====================================================\n")
    
    def full_information(self):
        if self.feedback == "counterfactual":
            self.m_th = 1.0
        elif self.feedback == "traj-Q":
            self.m_th = 1.0 / self.reach_prob
        else:
            self.m_th = self.opponent_reach_prob
        
        self.eta_tau = self.eta_coef + 1
        gradient = LiteEFG.aggregate(self.ev, "sum") + self.utility

        self.ev.inplace(LiteEFG.dot(gradient, self.u))
        
        self._update(self.u, self.u, gradient / (self.m_th / self.eta * self.alpha))
    
    def outcome_sampling(self):
        self.m_th = 1.0 / self.reach_prob
        self.eta_tau = self.eta_coef + 1 # 1 + eta * tau / m_th
        gradient = LiteEFG.aggregate(self.ev, "sum") + self.utility

        self.ev.inplace(gradient.sum())

        self._update(self.u, self.u, gradient / (self.u / self.eta * self.alpha))

    def _update(self, upd_u, ref_u, gradient):
        if self.regularizer == "Euclidean":
            upd_u.inplace((ref_u + gradient) / self.eta_tau)
            upd_u.inplace(upd_u.project(distance="L2", gamma=self.gamma, mu=self.mu))
        else:
            upd_u.inplace((ref_u.log() + gradient) / self.eta_tau)
            upd_u.inplace(upd_u - upd_u.max())
            upd_u.inplace(upd_u.exp())
            upd_u.inplace(upd_u.project(distance="KL", gamma=self.gamma, mu=self.mu))
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        env.update(self.u)
    
    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.u
    
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--iter", type=int, default=100000)
    parser.add_argument("--print_freq", type=int, default=1000)

    parser.add_argument("--eta", help="learning rate", type=float, default=0.1)
    parser.add_argument("--tau", help="regularization coefficient", type=float, default=0.01)
    parser.add_argument("--gamma", type=float, default=0.01)
    parser.add_argument("--regularizer", type=str, choices=["Euclidean", "Entropy"], default="Euclidean")
    parser.add_argument("--feedback", type=str, choices=["Q", "traj-Q", "counterfactual", "Outcome"], default="Q")
    parser.add_argument("--weighted", help="weighted dilated regularizer or not", action="store_true")

    args = parser.parse_args()

    traverse_type = "Enumerate" if(args.feedback != "Outcome") else "Outcome"

    from utils import train
    train(graph(args.eta, args.tau, args.gamma, args.regularizer, args.feedback, args.weighted), traverse_type, "default", args.iter, args.print_freq, args.game)
    # do not need to update strategy for algorithms only need last-iterate
    # it will be faster since LiteEFG do not need to maintain the sequence-form strategy
    # update_strategy() will enumerate all infosets. However, when using outcome-sampling, at each iteration,
    # the algorithm will only update a trajectory of length height
    # Solution: lazy-update


#######################################################
# Balanced Follow the Regularized Leader (Balanced FTRL)
# Fiegel, Côme, Pierre Ménard, Tadashi Kozuno, Rémi Munos, Vianney Perchet, and Michal Valko. 
# "Adapting to game trees in zero-sum imperfect information games." 
# International Conference on Machine Learning (2023).
#######################################################

import LiteEFG

class graph(LiteEFG.Graph):
    def __init__(self, eta=0.001, gamma=0.0005):
        super().__init__()

        self.eta = eta
        self.gamma = gamma
        self.initialized = False

        with LiteEFG.backward(is_static=True):
            self.subtree_action_number = self.action_set_size.copy()

            self.subtree_action_number.inplace(LiteEFG.aggregate(self.subtree_action_number, "sum", object="children", player="self", padding=0))
            self.strategy = LiteEFG.normalize(self.subtree_action_number + 1, p_norm=1.0)
            self.parent_subtree_size = LiteEFG.const(1, 0.0)

            self.subtree_action_number_vector = self.subtree_action_number.copy()
            self.subtree_action_number.inplace(self.subtree_action_number.sum() + self.action_set_size)
            
        with LiteEFG.forward(is_static=True):
            self.zero = LiteEFG.const(size=1, val=0.0)
            self.is_root = LiteEFG.aggregate(self.zero, "sum", object="parent", player="self", padding=1.0)

        with LiteEFG.forward(color=1):
            expectation = LiteEFG.const(size=1, val=0.0)
            self.visit_prob = LiteEFG.const(size=1, val=1.0)
            self.Z = LiteEFG.const(size=1, val=0.0)

            self.visit_prob.inplace(LiteEFG.aggregate(self.visit_prob, "sum", object="parent", player="self", padding=1.0))
            self.parent_subtree_size.inplace(LiteEFG.aggregate(self.subtree_action_number_vector, "sum", object="parent", player="self", padding=0) *\
                                              (1.0 - self.is_root) + self.parent_subtree_size * self.is_root)
            self.visit_prob.inplace(self.visit_prob / self.parent_subtree_size * self.subtree_action_number)
        
        with LiteEFG.backward(color=0):
            gradient = LiteEFG.aggregate(expectation, aggregator="sum") + self.utility / (self.reach_prob * self.strategy + self.gamma / self.visit_prob) * self.eta

            logit_max = LiteEFG.max(gradient)
            self.Z.inplace(LiteEFG.log(LiteEFG.sum(self.strategy * LiteEFG.exp(gradient - logit_max))) + logit_max) # Z_h^t
            self._update(gradient, self.strategy, self.strategy)
            expectation.inplace(self.Z.copy())

        print("===============Graph is ready for Balanced FTRL===============")
        print("eta: %f, gamma: %f" % (self.eta, self.gamma))
        print("==============================================================\n")
    
    def _update(self, gradient, upd_u, ref_u):
        upd_u.inplace(ref_u.log() + gradient)
        upd_u.inplace(upd_u - upd_u.max())
        upd_u.inplace(upd_u.exp())
        upd_u.inplace(upd_u.project(distance="KL"))
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        if not self.initialized:
            self.initialized = True
            for i in range(1, 3):
                subtree_action_number = env.get_value(i, self.subtree_action_number)
                is_root = env.get_value(i, self.is_root)

                parent_subtree_size = [[0.0] for _ in range(len(is_root))]

                total_actions = 0
                for j in range(len(is_root)):
                    if is_root[j][1][0] > 0.5: # root, >0.5 to avoid numerical error
                        total_actions += subtree_action_number[j][1][0]
                for j in range(len(is_root)):
                    if is_root[j][1][0] > 0.5:
                        parent_subtree_size[j][0] = total_actions
                env.set_value(i, self.parent_subtree_size, parent_subtree_size)
                env.update(self.strategy, upd_color=[1], traverse_type="Enumerate")

        env.update(self.strategy, upd_color=[0])

    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.strategy

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--eta", help="learning rate", type=float, default=0.001)
    parser.add_argument("--gamma", help="IX parameter", type=float, default=0.0005)
    parser.add_argument("--iter", type=int, default=1000000)
    parser.add_argument("--print_freq", type=int, default=1000)

    args = parser.parse_args()

    from utils import train
    train(graph(args.eta, args.gamma), "Outcome", "avg-iterate", args.iter, args.print_freq, args.game, output_strategy=True)



def _import_all_modules():
    import os
    import importlib
    package_dir = os.path.dirname(__file__)

    for filename in os.listdir(package_dir):
        if filename.endswith('.py') and filename != '__init__.py' and filename != 'utils.py':
            module_name = f'.{filename[:-3]}'
            importlib.import_module(module_name, package=__name__)

_import_all_modules()


import LiteEFG as LiteEFG
from __future__ import annotations
__all__ = ['LiteEFG', 'graph']
class graph(LiteEFG._LiteEFG.Graph):
    def __init__(self, eta = 0.001, tau = 0.001, gamma = 0.001, regularizer: typing.Literal['Euclidean', 'Entropy'] = 'Entropy', feedback = 'Q', weighted = False, bidilated = True):
        ...
    def _full_information(self):
        ...
    def _outcome_sampling(self):
        ...
    def _update(self, upd_u, ref_u, gradient):
        ...
    def current_strategy(self) -> LiteEFG._LiteEFG.GraphNode:
        ...
    def update_graph(self, env: LiteEFG._LiteEFG.Environment) -> None:
        ...



import pyspiel
import os
import LiteEFG
import numpy as np
import pandas as pd
from open_spiel.python.policy import TabularPolicy
import typing

def NodeName(node):
    if isinstance(node, str):
        return node.replace("\n", "/")
    return node.serialize().replace("\n", "/").replace(" ", "_")

def InfosetName(node, idx):
    infoset = "pl%d_%d__"%(0, idx) + node.replace("\n", "/") if isinstance(node, str) \
                    else "pl%d_%d__"%(node.current_player()+1, idx) + node.information_state_string().replace("\n", "/")
    return infoset.replace(" ", "_")

class OpenSpielEnv(LiteEFG.FileEnv):
    def __init__(self, game: pyspiel.Game, traverse_type="Enumerate", regenerate=False):
        if not isinstance(game, pyspiel.Game):
            raise ValueError("game must be an instance of pyspiel.Game")
        
        if game.get_type().dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL:
            self.game = game
        elif game.get_type().dynamics == pyspiel.GameType.Dynamics.SIMULTANEOUS: 
            game = pyspiel.convert_to_turn_based(game)
            self.game = game
        else:
            raise ValueError("The game must be either sequential or simultaneous")

        policy = TabularPolicy(game)
        self.state_lookup = {}
        for k in policy.state_lookup.keys():
            infoset = InfosetName(k, 0)
            infoset = infoset[infoset.find('__')+2:]
            self.state_lookup[infoset] = policy.state_lookup[k]

        game_name = game.get_type().short_name
        infosets = {}
        num_infosets = [0 for _ in range(game.num_players())]
        queue = [game.new_initial_state()]

        game_full_name = game_name
        for k in game.get_parameters():
            game_full_name += "_%s=%s"%(k, game.get_parameters()[k])

        current_directory = os.path.expanduser('~')#os.path.dirname(os.path.abspath(__file__))
        #for i in range(3):
        #    current_directory = os.path.dirname(current_directory)
        os.makedirs(os.path.join(current_directory, "game_instances"), exist_ok=True)
        file_name = os.path.join(current_directory, "game_instances", game_full_name + ".openspiel")

        if os.path.exists(file_name) and not regenerate:
            super().__init__(file_name, traverse_type=traverse_type)
            return
        
        print("Generating %s.openspiel instance from OpenSpiel"%(game_full_name))
        file = open(file_name, "w")

        print("# %s instance with parameters:"%game_name, file=file)
        print("#", file=file)
        print("# Opt {", file=file)
        print("#     openspiel,", file=file)
        for k in game.get_parameters():
            print("#     %s: %s,"%(k, game.get_parameters()[k]), file=file)
        if "players" not in game.get_parameters():
            print("#     players: %d,"%game.num_players(), file=file)
        print("# }", file=file)
        print("#", file=file)

        while len(queue) > 0:
            node = queue.pop()
            
            if node.is_terminal():
                print("node %s leaf payoffs"%NodeName(node), end='', file=file)
                for player, reward in enumerate(node.returns()):
                    print(" %d=%f"%(player+1, reward), end='', file=file)
                print(file=file)
                continue

            if node.is_chance_node():
                print("node %s chance actions"%NodeName(node), end='', file=file)
                for action, prob in node.chance_outcomes():
                    child = node.clone()
                    child.apply_action(action)
                    queue.append(child)
                    print(" %s=%.8f"%(NodeName(child.serialize()), prob), end='', file=file)
                print(file=file)
            else:
                print("node %s player %d actions"%(NodeName(node), node.current_player()+1), end='', file=file)
                for action in node.legal_actions():
                    child = node.clone()
                    child.apply_action(action)
                    queue.append(child)
                    print(" %s"%NodeName(child.serialize()), end='', file=file)
                print(file=file)

                try:
                    infoset = node.information_state_string()
                except:
                    raise ValueError("The game %s does not have information state implemented by OpenSpiel \
                                        (typically such games are also too large to run tabular algorithms)"%game_name)
                if infoset not in infosets:
                    infosets[infoset] = [InfosetName(node, num_infosets[node.current_player()])]
                    num_infosets[node.current_player()] += 1
                infosets[infoset].append(node.serialize())

        for infoset in infosets:
            print("infoset %s nodes"%infosets[infoset][0], end='', file=file)
            for node in infosets[infoset][1:]:
                print(" %s"%NodeName(node), end='', file=file)
            print(file=file)
        
        file.close()
        super().__init__(file_name, traverse_type=traverse_type)
    
    def get_value(self, player: int, node: LiteEFG.GraphNode) -> typing.List[typing.Tuple[str, float]]:
        values = super().get_value(player, node)
        policy = TabularPolicy(self.game)
        ret = []
        for k, _ in values:
            idx = self.state_lookup[k[k.find('__')+2:]]
            infoset = list(policy.state_lookup.keys())[idx]
            ret.append((infoset, _))
        return ret

    def get_strategy(self, strategy_node: LiteEFG.GraphNode, type_name="default") -> typing.Tuple[TabularPolicy, typing.List[pd.DataFrame]]:
        df_list = []
        policy = TabularPolicy(self.game)

        for player in range(self.game.num_players()):
            #print("Player %d strategy"%player)
            df = pd.DataFrame(columns=["Infoset"] + [self.game.action_to_string(player, _) for _ in range(self.game.num_distinct_actions())])
            strategy = super().get_strategy(player+1, strategy_node, type_name)
            data_list = []
            for infoset, probs in strategy:
                idx = self.state_lookup[infoset[infoset.find('__')+2:]]
                policy.action_probability_array[idx][policy.legal_actions_mask[idx]>0.5] = probs
                new_row = pd.DataFrame([[list(policy.state_lookup.keys())[idx]] + list(policy.action_probability_array[idx])], columns=df.columns)
                data_list.append(new_row)
            import warnings
            with warnings.catch_warnings():
                warnings.simplefilter(action='ignore', category=FutureWarning)
                df = pd.concat([df]+data_list, ignore_index=True)
            df_list.append(df)
        
        #from open_spiel.python.algorithms import exploitability
        #expl = exploitability.exploitability(self.game, policy)
        #print("Exploitability: %f"%expl)
        return policy, df_list

    def set_value(self, player: int, node: LiteEFG.GraphNode, values: typing.List[typing.List]) -> None:
        super().set_value(player, node, values)

    def interact(self, policy: TabularPolicy, controlled_player=0, reveal_private=True, epochs=1000) -> None:

        print("\nYou are player: %d in %s (players indexed from 0 to %d)"%(controlled_player, self.game.get_type().short_name, self.game.num_players()-1))

        accumulated_payoff = np.zeros(self.game.num_players())
        for epoch in range(epochs):
            print("\n========== Epoch %d ==========\n"%epoch)
            s = self.game.new_initial_state()
            while not s.is_terminal():
                if s.is_chance_node():
                    chance_outcomes = s.chance_outcomes()
                    prob = [_[1] for _ in chance_outcomes]
                    actions = [_[0] for _ in chance_outcomes]
                    action = np.random.choice(actions, p=prob)
                    s.apply_action(action)
                elif s.current_player() == controlled_player:
                    print(s.information_state_string())
                    actions = s.legal_actions()
                    while True:
                        print("Valid Actions: ", end='')
                        for i, action in enumerate(actions):
                            print("%d="%action + self.game.action_to_string(controlled_player, action), end=" ")
                        action = input("\nYour Choice: ")
                        if int(action) not in actions:
                            print("Invalid action! Please choose from the valid actions.")
                            continue
                        break
                    s.apply_action(int(action))
                else:
                    probs = policy.action_probabilities(s)
                    action = np.random.choice(list(probs.keys()), p=list(probs.values()))
                    s.apply_action(action)
            
            print("\n========== Epoch %d Summary ==========\n"%epoch)
            if not reveal_private:
                print("Your last observation is: ", s.information_state_string(controlled_player))
            else:
                print("The final outcome of the game is:\n")
                print(s)
            print("You are player: %d (players indexed from 0 to %d)"%(controlled_player, self.game.num_players()-1))
            print("Your Payoff: ", s.returns()[controlled_player])
            accumulated_payoff += s.returns()
            print("Accumulated Payoff of Each Player: ", accumulated_payoff)
            print("Average Payoff of Each Player: ", accumulated_payoff / (epoch+1))

if __name__ == "__main__":
    env = OpenSpielEnv(pyspiel.load_game("liars_dice(dice_sides=6)"))
    import LiteEFG.baselines.CFR as CFR
    import LiteEFG.baselines.CFRplus as CFRplus

    graph = CFR.graph()
    env.set_graph(graph)
    for i in range(10000):
        graph.update_graph(env)
        env.update_strategy(graph.current_strategy())
    policy, _ = env.get_strategy(graph.current_strategy(), "avg-iterate")
    env.interact(policy)



#######################################################
# Discounted Counterfactual Regret Minimization (DCFR)
# Brown, Noam, and Tuomas Sandholm.
# "Solving imperfect-information games via discounted regret minimization." 
# Proceedings of the AAAI Conference on Artificial Intelligence (2019).
#######################################################

import LiteEFG

class graph(LiteEFG.Graph):
    def __init__(self, alpha=1.5, beta=0, gamma=2):
        super().__init__()

        self.alpha = alpha
        self.beta = beta
        self.gamma = gamma
        self.threshold = 10

        with LiteEFG.backward(is_static=True):

            self.timestep = LiteEFG.const(size=1, val=0.0)
            expectation = LiteEFG.const(size=1, val=0.0)
            self.strategy = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)
            self.avg_seq_strategy = LiteEFG.const(self.action_set_size, 0.0)
            self.regret_buffer = LiteEFG.const(self.action_set_size, 0.0)
            self.avg_strategy = self.strategy.copy()

            self.pos_coeff = LiteEFG.const(1, 1.0) if self.alpha>self.threshold else LiteEFG.const(1, 0.0)
            self.neg_coeff = LiteEFG.const(1, 1.0) if self.beta>self.threshold else LiteEFG.const(1, 0.0)

        with LiteEFG.backward():

            self.strategy_coef = (self.timestep / (self.timestep + 1)) ** self.gamma
            self.timestep.inplace(self.timestep + 1)
            
            counterfactual_value = LiteEFG.aggregate(expectation, aggregator="sum") + self.utility
            expectation.inplace(LiteEFG.dot(counterfactual_value, self.strategy))

            self.neg_regret = self.regret_buffer < 0
            self.pos_regret = self.regret_buffer >= 0

            self.regret_buffer.inplace((self.neg_regret * self.regret_buffer) * self.neg_coeff + (self.pos_regret * self.regret_buffer) * self.pos_coeff)

            self.regret_buffer.inplace(self.regret_buffer + counterfactual_value - expectation)
            self.avg_seq_strategy.inplace(self.avg_seq_strategy * self.strategy_coef + self.strategy * self.reach_prob)
            self.avg_strategy.inplace(self.avg_seq_strategy.normalize(p_norm=1.0, ignore_negative=True))
            self.strategy.inplace(LiteEFG.normalize(self.regret_buffer, p_norm=1.0, ignore_negative=True))

            if abs(self.alpha) < self.threshold:
                self.pos_coeff.inplace((self.timestep ** self.alpha) / (self.timestep ** self.alpha + 1))
            if abs(self.beta) < self.threshold:
                self.neg_coeff.inplace((self.timestep ** self.beta) / (self.timestep ** self.beta + 1))
            
        print("===============Graph is ready for DCFR===============")
        print("alpha: %f, beta: %f, gamma: %f" % (self.alpha, self.beta, self.gamma))
        print("====================================================\n")

    def update_graph(self, env : LiteEFG.Environment) -> None:
        env.update(self.strategy, upd_player=1)
        env.update(self.strategy, upd_player=2)
    
    def current_strategy(self, type_name="last-iterate"):
        assert(type_name in ["last-iterate", "average-iterate"])
        return self.strategy if type_name == "average-iterate" else self.avg_strategy

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--traverse_type", type=str, choices=["Enumerate", "External"], default="Enumerate")
    parser.add_argument("--alpha", type=float, default=1.5)
    parser.add_argument("--beta", type=float, default=0)
    parser.add_argument("--gamma", type=float, default=2)
    parser.add_argument("--iter", type=int, default=100000)
    parser.add_argument("--print_freq", type=int, default=1000)

    args = parser.parse_args()
    from utils import train
    train(graph(args.alpha, args.beta, args.gamma), args.traverse_type, "last-iterate", args.iter, args.print_freq, args.game)



#######################################################
# Counterfactual Regret Minimization+ (CFR+)
# Zinkevich, Martin, Michael Johanson, Michael Bowling, and Carmelo Piccione.
# "Regret minimization in games with incomplete information." 
# Advances in neural information processing systems (2007).
#######################################################

import LiteEFG

class graph(LiteEFG.Graph):
    def __init__(self):
        super().__init__()
        
        with LiteEFG.backward(is_static=True):
                
            expectation = LiteEFG.const(1, 0.0)
            self.strategy = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)
            self.regret_buffer = LiteEFG.const(self.action_set_size, 0.0)

        # RM+
        with LiteEFG.backward():

            counterfactual_value = LiteEFG.aggregate(expectation, aggregator="sum") + self.utility
            expectation.inplace(LiteEFG.dot(counterfactual_value, self.strategy))
            self.regret_buffer.inplace(LiteEFG.maximum(self.regret_buffer + counterfactual_value - expectation, 0.0))
            self.strategy.inplace(LiteEFG.normalize(self.regret_buffer, p_norm=1.0, ignore_negative=True))

        print("===============Graph is ready for CFR+===============")
        print()
        print("====================================================\n")
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        env.update(self.strategy, upd_player=1)
        env.update(self.strategy, upd_player=2)
    
    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.strategy
    
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--traverse_type", type=str, choices=["Enumerate", "External"], default="Enumerate")
    parser.add_argument("--iter", type=int, default=100000)
    parser.add_argument("--print_freq", type=int, default=1000)

    args = parser.parse_args()

    from utils import train
    train(graph(), args.traverse_type, "linear-avg-iterate", args.iter, args.print_freq, args.game)



#######################################################
# Outcome Sampling Monte-Carlo Counterfactual Regret Minimization (OS-MCCFR)
# Lanctot, Marc, Kevin Waugh, Martin Zinkevich, and Michael Bowling.
# "Monte Carlo sampling for regret minimization in extensive games." 
# Advances in neural information processing systems (2009).
#######################################################

import LiteEFG

class graph(LiteEFG.Graph):
    def __init__(self, delta=0.1, rm_plus=False, balanced=True):
        super().__init__()

        self.delta = delta
        self.rm_plus = rm_plus

        with LiteEFG.forward(is_static=True):

            expectation = LiteEFG.const(size=1, val=0.0)
            self.strategy = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)
            self.sequence_form_strategy = self.strategy.copy()
            self.prev_strategy = self.strategy.copy()
            
            self.strategy_reach = LiteEFG.aggregate(self.sequence_form_strategy, aggregator="sum", object="parent", padding=1.0)
            # return pad when the object not found. Here it is return 1.0 when no parent exists (root infoset)
            self.sequence_form_strategy.inplace(self.strategy_reach * self.strategy)

            if balanced:
                self.uniform = LiteEFG.normalize(self.subtree_size, p_norm=1.0, ignore_negative=True)
            else:
                self.uniform = self.strategy.copy()
            self.sequence_form_uniform = self.uniform.copy()
            self.explore_strategy = self.uniform.copy()
        
            self.uniform_reach = LiteEFG.aggregate(self.sequence_form_uniform, aggregator="sum", object="parent", padding=1.0)
            self.sequence_form_uniform.inplace(self.uniform_reach * self.uniform)

            self.regret_buffer = LiteEFG.const(self.action_set_size, 0.0)

        with LiteEFG.backward():

            counterfactual_value = LiteEFG.aggregate(expectation, aggregator="sum")
            counterfactual_value.inplace(counterfactual_value + self.utility / (self.reach_prob * self.explore_strategy))
            expectation.inplace(LiteEFG.dot(counterfactual_value, self.strategy))

            self.regret_buffer.inplace(self.regret_buffer + counterfactual_value - expectation)
            self.prev_strategy.inplace(self.strategy.copy())
            self.strategy.inplace(LiteEFG.normalize(self.regret_buffer, p_norm=1.0, ignore_negative=True))
            if rm_plus:
                self.regret_buffer.inplace(LiteEFG.maximum(self.regret_buffer, 0.0))

        with LiteEFG.forward():

            self.strategy_reach.inplace(LiteEFG.aggregate(self.sequence_form_strategy, aggregator="sum", object="parent", padding=1.0))
            self.sequence_form_strategy.inplace(self.strategy_reach * self.strategy)
            self.explore_strategy.inplace(LiteEFG.normalize(delta * self.sequence_form_uniform + (1 - delta) * self.sequence_form_strategy, p_norm=1.0))

        print("===============Graph is ready for Outcome-Sampling MCCFR===============")
        print("delta: %f, rm_plus: %s" % (self.delta, self.rm_plus))
        print("====================================================\n")
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        env.update([self.explore_strategy, self.strategy], upd_player=1)
        env.update([self.prev_strategy, self.explore_strategy], upd_player=2)
    
    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.strategy

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--rm+", dest="rm_plus", help="regret matching+ or not", action="store_true")
    parser.add_argument("--balanced", help="balance exploration or not", action="store_true")
    parser.add_argument("--delta", help="exploration rate", type=float, default=0.1)
    parser.add_argument("--iter", type=int, default=100000)
    parser.add_argument("--print_freq", type=int, default=1000)

    args = parser.parse_args()

    from utils import train
    train(graph(args.delta, args.rm_plus, args.balanced), "Outcome", "avg-iterate", args.iter, args.print_freq, args.game, output_strategy=True)



from LiteEFG._LiteEFG import Environment
from LiteEFG._LiteEFG import FileEnv
from LiteEFG._LiteEFG import Graph
from LiteEFG._LiteEFG import GraphNode
from LiteEFG._LiteEFG import GraphNodeStatus
from LiteEFG._LiteEFG import Vector
from LiteEFG._LiteEFG import aggregate
from LiteEFG._LiteEFG import argmax
from LiteEFG._LiteEFG import argmin
from LiteEFG._LiteEFG import backward
from LiteEFG._LiteEFG import cat
from LiteEFG._LiteEFG import const
from LiteEFG._LiteEFG import copy
from LiteEFG._LiteEFG import dot
from LiteEFG._LiteEFG import euclidean
from LiteEFG._LiteEFG import exp
from LiteEFG._LiteEFG import forward
from LiteEFG._LiteEFG import log
from LiteEFG._LiteEFG import max
from LiteEFG._LiteEFG import maximum
from LiteEFG._LiteEFG import mean
from LiteEFG._LiteEFG import min
from LiteEFG._LiteEFG import minimum
from LiteEFG._LiteEFG import negative_entropy
from LiteEFG._LiteEFG import normalize
from LiteEFG._LiteEFG import project
from LiteEFG._LiteEFG import set_seed
from LiteEFG._LiteEFG import sum
from LiteEFG.src.Environment.OpenSpiel.OpenSpielToGameFile import OpenSpielEnv
from __future__ import annotations
from . import _LiteEFG
from . import baselines
from . import src
__all__ = ['Environment', 'FileEnv', 'Graph', 'GraphNode', 'GraphNodeStatus', 'OpenSpielEnv', 'Vector', 'aggregate', 'argmax', 'argmin', 'backward', 'baselines', 'cat', 'const', 'copy', 'dot', 'euclidean', 'exp', 'forward', 'log', 'max', 'maximum', 'mean', 'min', 'minimum', 'negative_entropy', 'normalize', 'project', 'set_seed', 'src', 'sum']



from __future__ import annotations
import typing
__all__ = ['Environment', 'FileEnv', 'Graph', 'GraphNode', 'GraphNodeStatus', 'Vector', 'aggregate', 'argmax', 'argmin', 'backward', 'cat', 'const', 'copy', 'dot', 'euclidean', 'exp', 'forward', 'log', 'max', 'maximum', 'mean', 'min', 'minimum', 'negative_entropy', 'normalize', 'project', 'set_seed', 'sum']
class Environment:
    @typing.overload
    def exploitability(self, strategy: GraphNode, type_name: str = 'default') -> list[float]:
        ...
    @typing.overload
    def exploitability(self, strategy: list[GraphNode], type_name: str = 'default') -> list[float]:
        ...
    def get_strategy(self, player: int, strategy: GraphNode, type_name: str = 'default') -> list[tuple[str, list[float]]]:
        ...
    def get_value(self, player: int, node: GraphNode) -> list[tuple[str, list[float]]]:
        ...
    def set_graph(self, graph: Graph) -> None:
        ...
    def set_value(self, player: int, node: GraphNode, values: list[list[float]]) -> None:
        ...
    @typing.overload
    def update(self, strategy: GraphNode, upd_player: int = -1, upd_color: list[int] = [-1], traverse_type: str = 'default') -> None:
        ...
    @typing.overload
    def update(self, strategies: list[GraphNode], upd_player: int = -1, upd_color: list[int] = [-1], traverse_type: str = 'default') -> None:
        ...
    def update_strategy(self, strategy: GraphNode, update_best: bool = False) -> None:
        ...
    @typing.overload
    def utility(self, strategy: GraphNode, type_name: str = 'default') -> list[float]:
        ...
    @typing.overload
    def utility(self, strategy: list[GraphNode], type_name: str = 'default') -> list[float]:
        ...
class FileEnv(Environment):
    def __init__(self, file_name: str, traverse_type: str = 'Enumerate') -> None:
        ...
class Graph:
    def __init__(self) -> None:
        ...
    @property
    def action_set_size(self) -> GraphNode:
        ...
    @property
    def opponent_reach_prob(self) -> GraphNode:
        ...
    @property
    def reach_prob(self) -> GraphNode:
        ...
    @property
    def subtree_size(self) -> GraphNode:
        ...
    @property
    def utility(self) -> GraphNode:
        ...
class GraphNode:
    __hash__: typing.ClassVar[None] = None
    def __add__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    @typing.overload
    def __eq__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    @typing.overload
    def __eq__(self, arg0: float | int) -> GraphNode:
        ...
    @typing.overload
    def __ge__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    @typing.overload
    def __ge__(self, arg0: float | int) -> GraphNode:
        ...
    @typing.overload
    def __gt__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    @typing.overload
    def __gt__(self, arg0: float | int) -> GraphNode:
        ...
    def __init__(self) -> None:
        ...
    @typing.overload
    def __le__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    @typing.overload
    def __le__(self, arg0: float | int) -> GraphNode:
        ...
    @typing.overload
    def __lt__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    @typing.overload
    def __lt__(self, arg0: float | int) -> GraphNode:
        ...
    def __mul__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    def __neg__(self) -> GraphNode:
        ...
    def __pow__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    def __radd__(self, arg0: float | int) -> GraphNode:
        ...
    def __rmul__(self, arg0: float | int) -> GraphNode:
        ...
    def __rpow__(self: float | int, arg0: GraphNode) -> GraphNode:
        ...
    def __rsub__(self, arg0: float | int) -> GraphNode:
        ...
    def __rtruediv__(self, arg0: float | int) -> GraphNode:
        ...
    def __sub__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    def __truediv__(self, arg0: int | float | GraphNode) -> GraphNode:
        ...
    def argmax(self) -> GraphNode:
        ...
    def argmin(self) -> GraphNode:
        ...
    def copy(self) -> GraphNode:
        ...
    def dot(self, arg0: GraphNode) -> GraphNode:
        ...
    def euclidean(self) -> GraphNode:
        ...
    def exp(self) -> GraphNode:
        ...
    def inplace(self, arg0: GraphNode) -> None:
        ...
    def log(self) -> GraphNode:
        ...
    def max(self) -> GraphNode:
        ...
    def mean(self) -> GraphNode:
        ...
    def min(self) -> GraphNode:
        ...
    def negative_entropy(self, shifted: bool = False) -> GraphNode:
        ...
    def normalize(self, p_norm: float, ignore_negative: bool = False) -> GraphNode:
        ...
    @typing.overload
    def project(self, distance: str, gamma: int | float | GraphNode = 0.0) -> GraphNode:
        ...
    @typing.overload
    def project(self, distance: str, gamma: int | float | GraphNode, mu: GraphNode) -> GraphNode:
        ...
    def sum(self) -> GraphNode:
        ...
class GraphNodeStatus:
    def __enter__(self) -> GraphNodeStatus:
        ...
    def __exit__(self, *args) -> None:
        ...
    def __init__(self) -> None:
        ...
class Vector:
    def __init__(self) -> None:
        ...
    def __repr__(self) -> str:
        ...
    def print(self) -> None:
        ...
class backward(GraphNodeStatus):
    def __init__(self, is_static: bool = False, color: int = 0) -> None:
        ...
class forward(GraphNodeStatus):
    def __init__(self, is_static: bool = False, color: int = 0) -> None:
        ...
def aggregate(arg0: GraphNode, aggregator: str, object: str = 'children', player: str = 'self', padding: float = 0.0) -> GraphNode:
    ...
def argmax(arg0: GraphNode) -> GraphNode:
    ...
def argmin(arg0: GraphNode) -> GraphNode:
    ...
def cat(nodes: list[GraphNode]) -> GraphNode:
    ...
def const(size: int | float | GraphNode, val: int | float | GraphNode) -> GraphNode:
    ...
def copy(arg0: GraphNode) -> GraphNode:
    ...
def dot(arg0: GraphNode, arg1: GraphNode) -> GraphNode:
    ...
def euclidean(arg0: GraphNode) -> GraphNode:
    ...
def exp(arg0: GraphNode) -> GraphNode:
    ...
def log(arg0: GraphNode) -> GraphNode:
    ...
def max(arg0: GraphNode) -> GraphNode:
    ...
def maximum(arg0: GraphNode, arg1: int | float | GraphNode) -> GraphNode:
    ...
def mean(arg0: GraphNode) -> GraphNode:
    ...
def min(arg0: GraphNode) -> GraphNode:
    ...
def minimum(arg0: GraphNode, arg1: int | float | GraphNode) -> GraphNode:
    ...
def negative_entropy(arg0: GraphNode, shifted: bool = False) -> GraphNode:
    ...
def normalize(arg0: GraphNode, p_norm: float, ignore_negative: bool = False) -> GraphNode:
    ...
@typing.overload
def project(arg0: GraphNode, distance: str, gamma: int | float | GraphNode = 0.0) -> GraphNode:
    ...
@typing.overload
def project(arg0: GraphNode, distance: str, gamma: int | float | GraphNode, mu: GraphNode) -> GraphNode:
    ...
def set_seed(seed: int) -> None:
    ...
def sum(arg0: GraphNode) -> GraphNode:
    ...



#######################################################
# Implicit Exploration Online Mirror Descent (IXOMD)
# Kozuno, Tadashi, Pierre Ménard, Rémi Munos, and Michal Valko.
# "Learning in two-player zero-sum partially observable Markov games with perfect recall."
# Advances in Neural Information Processing Systems (2021).
#######################################################

import LiteEFG

class graph(LiteEFG.Graph):
    def __init__(self, eta=0.001, gamma=0.0005):
        super().__init__()

        self.eta = eta
        self.gamma = gamma
        self.initialized = False

        with LiteEFG.forward(is_static=True):
            self.strategy = LiteEFG.const(self.action_set_size, 1.0/self.action_set_size)
            expectation = LiteEFG.const(size=1, val=0.0)
            self.Z = LiteEFG.const(size=1, val=1.0)
        
        with LiteEFG.backward():
            gradient = LiteEFG.aggregate(expectation, aggregator="sum") + self.utility / (self.reach_prob * self.strategy + self.gamma) * self.eta
            # gradient = \eta * r_h^t + \log Z_{h+1}^t

            logit_max = LiteEFG.max(gradient)
            self.Z.inplace(LiteEFG.log(LiteEFG.sum(self.strategy * LiteEFG.exp(gradient - logit_max))) + logit_max) # Z_h^t
            self._update(gradient, self.strategy, self.strategy)
            expectation.inplace(self.Z.copy())

        print("===============Graph is ready for IXOMD===============")
        print("eta: %f, gamma: %f" % (self.eta, self.gamma))
        print("======================================================\n")
    
    def _update(self, gradient, upd_u, ref_u):
        upd_u.inplace(ref_u.log() + gradient)
        upd_u.inplace(upd_u - upd_u.max())
        upd_u.inplace(upd_u.exp())
        upd_u.inplace(upd_u.project(distance="KL"))
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        env.update(self.strategy)

    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.strategy

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--eta", help="learning rate", type=float, default=0.001)
    parser.add_argument("--gamma", help="IX parameter", type=float, default=0.0005)
    parser.add_argument("--iter", type=int, default=1000000)
    parser.add_argument("--print_freq", type=int, default=1000)

    args = parser.parse_args()

    from utils import train
    train(graph(args.eta, args.gamma), "Outcome", "avg-iterate", args.iter, args.print_freq, args.game, output_strategy=True)



from __future__ import annotations
from . import Balanced_FTRL
from . import Balanced_OMD
from . import CFR
from . import CFRplus
from . import CMD
from . import DCFR
from . import DOMD
from . import IXOMD
from . import MMD
from . import OS_MCCFR
from . import PCFR
from . import QFR
from . import Reg_CFR
from . import Reg_DOMD
__all__ = ['Balanced_FTRL', 'Balanced_OMD', 'CFR', 'CFRplus', 'CMD', 'DCFR', 'DOMD', 'IXOMD', 'MMD', 'OS_MCCFR', 'PCFR', 'QFR', 'Reg_CFR', 'Reg_DOMD']
def _import_all_modules():
    ...



import LiteEFG as LiteEFG
from __future__ import annotations
__all__ = ['LiteEFG', 'graph']
class graph(LiteEFG._LiteEFG.Graph):
    def __init__(self, delta = 0.1, rm_plus = False, balanced = True):
        ...
    def current_strategy(self) -> LiteEFG._LiteEFG.GraphNode:
        ...
    def update_graph(self, env: LiteEFG._LiteEFG.Environment) -> None:
        ...



#######################################################
# Regularized Dilated Optimistic Mirror Descent (Reg-DOMD)
# Mingyang Liu, Asuman Ozdaglar, Tiancheng Yu, and Kaiqing Zhang.
# "The power of regularization in solving extensive-form games."
# International Conference on Learning Representations (2023).
#######################################################

import LiteEFG
from typing import Literal

class graph(LiteEFG.Graph):
    def __init__(self, eta=0.1, tau=0.1, regularizer: Literal["Euclidean", "Entropy"]="Entropy", 
                        weighted=False, shrink_iter=100000, out_reg=False):
        super().__init__()
        self.eta = eta
        self.regularizer = regularizer
        self.timestep = 0
        self.shrink_iter = shrink_iter
        self.out_reg = out_reg

        # Create a new graph for CFR
        with LiteEFG.backward(is_static=True):
        
            self.alpha = 1.0
            if weighted:
                self.alpha = LiteEFG.const(1, 1.0)
                self.alpha.inplace(LiteEFG.aggregate(self.alpha, "sum"))
                self.alpha.inplace((self.alpha.max() + 1) * 2)

            ev = LiteEFG.const(size=1, val=0.0)
            bar_ev = LiteEFG.const(size=1, val=0.0)

            self.tau = LiteEFG.const(1, tau)
            self.eta_coef = self.alpha / eta
            self.coef = self.tau
            self.u = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)
            self.bar_u = self.u.copy()

        with LiteEFG.backward(color=0):

            bar_gradient = LiteEFG.aggregate(bar_ev, "sum") + self.utility
            gradient = LiteEFG.aggregate(ev, "sum") + self.utility

            self.prev_bar_u =  self.bar_u.copy()
            self._update(bar_gradient, self.bar_u, self.bar_u)
            self._update(gradient, self.u, self.bar_u)

            self._get_ev(bar_gradient, bar_ev, self.bar_u, self.prev_bar_u)
            self._get_ev(gradient, ev, self.u, self.bar_u)
        
        with LiteEFG.backward(color=1):
            self.tau.inplace(self.tau * 0.5)
            self.coef.inplace(self.tau)

        print("===============Graph is ready for Reg-DOMD===============")
        print("eta: %f, tau: %f, regularizer: %s" % (self.eta, tau, self.regularizer))
        print("=========================================================\n")
    
    def _get_ev(self, gradient, ev, strategy, ref_strategy):
        if not self.out_reg:
            if self.regularizer == "Euclidean":
                ev.inplace(LiteEFG.dot(gradient, strategy) - LiteEFG.euclidean(strategy - ref_strategy) * self.eta_coef
                                                            - (LiteEFG.dot(ref_strategy, strategy)) * self.tau * self.alpha)
            else:
                kl = LiteEFG.dot((strategy / ref_strategy).log(), strategy) * self.eta_coef
                ev.inplace(LiteEFG.dot(gradient, strategy) - kl - (LiteEFG.dot(ref_strategy.log(), strategy)) * self.tau * self.alpha)
        else:
            if self.regularizer == "Euclidean":
                ev.inplace(LiteEFG.dot(gradient, strategy) - LiteEFG.euclidean(strategy - ref_strategy) * self.eta_coef
                                                            - LiteEFG.euclidean(strategy) * self.tau * self.alpha)
            else:
                kl = LiteEFG.dot((strategy / ref_strategy).log(), strategy) * self.eta_coef
                ev.inplace(LiteEFG.dot(gradient, strategy) - kl - LiteEFG.negative_entropy(strategy) * self.tau * self.alpha)

    def _update(self, gradient, upd_u, ref_u):
        if not self.out_reg:
            if self.regularizer == "Euclidean":
                gradient.inplace(gradient - ref_u * self.alpha * self.tau)
                gradient_div = gradient / self.eta_coef
                upd_u.inplace(ref_u + gradient_div)
                upd_u.inplace(upd_u.project(distance="L2"))

            else:
                gradient.inplace(gradient - (ref_u.log() + 1) * self.alpha * self.tau)
                gradient_div = gradient / self.eta_coef
                upd_u.inplace(ref_u.log() + gradient_div)
                upd_u.inplace(upd_u - upd_u.max())
                upd_u.inplace(upd_u.exp())
                upd_u.inplace(upd_u.project(distance="KL"))
        else:
            if self.regularizer == "Euclidean":
                gradient_div = gradient / self.eta_coef
                upd_u.inplace((ref_u + gradient_div) / (1.0 + self.coef * self.eta))
                upd_u.inplace(upd_u.project(distance="L2"))

            else:
                gradient_div = gradient / self.eta_coef
                upd_u.inplace((ref_u.log() + gradient_div) / (1.0 + self.coef * self.eta))
                upd_u.inplace(upd_u - upd_u.max())
                upd_u.inplace(upd_u.exp())
                upd_u.inplace(upd_u.project(distance="KL"))
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        self.timestep += 1
        env.update(self.u, upd_color=[0, 1]) if self.timestep % self.shrink_iter == 0 else env.update(self.u, upd_color=[0])
    
    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.u
    
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--traverse_type", type=str, choices=["Enumerate", "External", "Outcome"], default="Enumerate")
    parser.add_argument("--iter", type=int, default=100000)
    parser.add_argument("--print_freq", type=int, default=1000)

    parser.add_argument("--eta", help="learning rate", type=float, default=0.1)
    parser.add_argument("--tau", help="regularization coefficient", type=float, default=0.001)
    parser.add_argument("--shrink-iter", help="shrink tau by half every shrink-iter iterations", type=int, default=100000)
    parser.add_argument("--regularizer", type=str, choices=["Euclidean", "Entropy"], default="Euclidean")
    parser.add_argument("--weighted", help="weighted dilated regularizer or not", action="store_true")
    parser.add_argument("--out-reg", help="whether the update-rule is \\argmin <g, x+\\nabla\\psi(x_0)+D(x, x_0)> (default) or \\argmin <g, x>+\\psi(x)+D(x, x_0)", action="store_true")

    args = parser.parse_args()

    from utils import train
    train(graph(args.eta, args.tau, args.regularizer, args.weighted, args.shrink_iter, args.out_reg), args.traverse_type, "last-iterate", args.iter, args.print_freq, args.game)



#######################################################
# Clairvoyant Mirror Descent (CMD)
# Piliouras, Georgios, Ryann Sim, and Stratis Skoulakis.
# "Beyond time-average convergence: Near-optimal uncoupled online learning via clairvoyant multiplicative weights update." 
# Advances in Neural Information Processing Systems (2022).
#######################################################

import LiteEFG
from typing import Literal

class graph(LiteEFG.Graph):
    def __init__(self, eta=0.1, inner_epoch=10, regularizer: Literal["Euclidean", "Entropy"]="Entropy", weighted=False):
        super().__init__()
        self.eta = eta
        self.regularizer = regularizer
        self.timestep = 0
        self.inner_epoch = inner_epoch

        # Create a new graph for CFR
        with LiteEFG.backward(is_static=True):
        
            self.alpha = 1.0
            if weighted:
                self.alpha = LiteEFG.const(1, 1.0)
                self.alpha.inplace(LiteEFG.aggregate(self.alpha, "sum"))
                self.alpha.inplace((self.alpha.max() + 1) * 2)

            ev = LiteEFG.const(size=1, val=0.0)

            self.coef = self.alpha / eta
            self.u = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)
            self.bar_u = self.u.copy()

        with LiteEFG.backward():

            gradient = LiteEFG.aggregate(ev, "sum") + self.utility

            self._update(gradient, self.u, self.bar_u)
            self._get_ev(gradient, ev, self.u, self.bar_u)
        
        with LiteEFG.backward(color=1):
            
            self.bar_u.inplace(self.u.copy())

        print("===============Graph is ready for CMD===============")
        print("eta: %f, inner epoch: %d, regularizer: %s" % (self.eta, self.inner_epoch, self.regularizer))
        print("=====================================================\n")
    
    def _get_ev(self, gradient, ev, strategy, ref_strategy):
        if self.regularizer == "Euclidean":
            ev.inplace(LiteEFG.dot(gradient, strategy) - LiteEFG.euclidean(strategy - ref_strategy) * self.coef)
        else:
            kl = LiteEFG.dot((strategy / ref_strategy).log(), strategy) * self.coef
            ev.inplace(LiteEFG.dot(gradient, strategy) - kl)

    def _update(self, gradient, upd_u, ref_u):
        gradient_div = gradient / self.coef
        
        if self.regularizer == "Euclidean":
            upd_u.inplace(ref_u + gradient_div)
            upd_u.inplace(upd_u.project(distance="L2"))

        else:
            upd_u.inplace(ref_u.log() + gradient_div)
            upd_u.inplace(upd_u - upd_u.max())
            upd_u.inplace(upd_u.exp())
            upd_u.inplace(upd_u.project(distance="KL"))
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        self.timestep += 1
        if self.timestep % self.inner_epoch == 0:
            env.update(self.u, upd_color=[0, 1])
        else:
            env.update(self.u, upd_color=[0])

    
    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.bar_u
    
if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--traverse_type", type=str, choices=["Enumerate", "External", "Outcome"], default="Enumerate")
    parser.add_argument("--iter", type=int, default=100000)
    parser.add_argument("--print_freq", type=int, default=1000)

    parser.add_argument("--eta", help="learning rate", type=float, default=0.1)
    parser.add_argument("--inner_epoch", type=int, default=10)
    parser.add_argument("--regularizer", type=str, choices=["Euclidean", "Entropy"], default="Euclidean")
    parser.add_argument("--weighted", help="weighted dilated regularizer or not", action="store_true")

    args = parser.parse_args()

    from utils import train
    train(graph(args.eta, args.inner_epoch, args.regularizer, args.weighted), args.traverse_type, "last-iterate", args.iter, args.print_freq, args.game)



#######################################################
# Balanced Online Mirror Descent (Balanced-OMD)
# Bai, Yu, Chi Jin, Song Mei, and Tiancheng Yu.
# "Near-optimal learning of extensive-form games with imperfect information."
# International Conference on Machine Learning (2022).
#######################################################

import LiteEFG
import numpy as np

class graph(LiteEFG.Graph):
    def __init__(self, eta=0.001, gamma=0.0005):
        super().__init__()

        self.eta = eta
        self.gamma = gamma
        self.initialized = False

        with LiteEFG.forward(is_static=True):
            self.depth = LiteEFG.const(size=1, val=1.0)
            self.depth.inplace(LiteEFG.aggregate(self.depth, "max", object="parent", player="self", padding=0) + 1.0)
            self.strategy = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)
            expectation = LiteEFG.const(size=1, val=0.0)
            self.visit_prob = LiteEFG.const(size=1, val=0.0)
        
        with LiteEFG.backward():
            self.visit_action_prob = self.visit_prob / self.action_set_size
            gradient = LiteEFG.aggregate(expectation, aggregator="sum") + self.utility / (self.reach_prob * self.strategy + self.gamma * self.visit_action_prob)

            self.prev_strategy =  self.strategy.copy()
            self._update(gradient, self.strategy, self.prev_strategy)

            kl = LiteEFG.dot((self.strategy / self.prev_strategy).log(), self.strategy) / (self.eta * self.visit_action_prob)
            expectation.inplace(LiteEFG.dot(gradient, self.strategy) - kl)

        print("===============Graph is ready for Balanced OMD===============")
        print("eta: %f, gamma: %f" % (self.eta, self.gamma))
        print("=============================================================\n")
    
    def _update(self, gradient, upd_u, ref_u):
        gradient_div = gradient * self.eta * self.visit_action_prob
        upd_u.inplace(ref_u.log() + gradient_div)
        upd_u.inplace(upd_u - upd_u.max())
        upd_u.inplace(upd_u.exp())
        upd_u.inplace(upd_u.project(distance="KL"))
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        if not self.initialized:
            self.initialized = True
            for i in range(1, 3): # compute \mu^{*,h}_{1:h} for each infoset
                depth = env.get_value(i, self.depth)
                max_depth = 0
                for j in depth:
                    max_depth = max(max_depth, np.round(j[1][0]).astype(int))
                depth_infoset_count = np.zeros(max_depth + 1)

                for j in depth:
                    depth_infoset_count[np.round(j[1][0]).astype(int)] += 1
                
                visit_prob = [[0] for _ in range(len(depth))]
                for j in range(len(depth)):
                    visit_prob[j][0] = 1.0# / depth_infoset_count[np.round(depth[j][1][0]).astype(int)]
                env.set_value(i, self.visit_prob, visit_prob)
            
        env.update(self.strategy)

    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.strategy

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker")
    parser.add_argument("--eta", help="learning rate", type=float, default=0.001)
    parser.add_argument("--gamma", help="IX parameter", type=float, default=0.0005)
    parser.add_argument("--iter", type=int, default=1000000)
    parser.add_argument("--print_freq", type=int, default=1000)

    args = parser.parse_args()

    from utils import train
    train(graph(args.eta, args.gamma), "Outcome", "avg-iterate", args.iter, args.print_freq, args.game, output_strategy=True)



import LiteEFG as LiteEFG
from __future__ import annotations
import numpy as np
import open_spiel.python.policy
from open_spiel.python.policy import TabularPolicy
import os as os
import pandas as pd
import pyspiel as pyspiel
import typing as typing
__all__ = ['InfosetName', 'LiteEFG', 'NodeName', 'OpenSpielEnv', 'TabularPolicy', 'np', 'os', 'pd', 'pyspiel', 'typing']
class OpenSpielEnv(LiteEFG._LiteEFG.FileEnv):
    def __init__(self, game: pyspiel.Game, traverse_type = 'Enumerate', regenerate = False):
        ...
    def get_strategy(self, strategy_node: LiteEFG._LiteEFG.GraphNode, type_name = 'default') -> typing.Tuple[open_spiel.python.policy.TabularPolicy, typing.List[pandas.core.frame.DataFrame]]:
        ...
    def get_value(self, player: int, node: LiteEFG._LiteEFG.GraphNode) -> typing.List[typing.Tuple[str, float]]:
        ...
    def interact(self, policy: open_spiel.python.policy.TabularPolicy, controlled_player = 0, reveal_private = True, epochs = 1000) -> None:
        ...
    def set_value(self, player: int, node: LiteEFG._LiteEFG.GraphNode, values: typing.List[typing.List]) -> None:
        ...
def InfosetName(node, idx):
    ...
def NodeName(node):
    ...



#######################################################
# Q-Function based Regret Minimization (QFR)
# Mingyang Liu, Gabriele Farina, and Asuman Ozdaglar.
# "A Policy-Gradient Approach to Solving Imperfect-Information Games with Iterate Convergence."
#######################################################

import  LiteEFG
from typing import Literal

class graph(LiteEFG.Graph):
    def __init__(self, eta=0.001, tau=0.001, gamma=0.001, regularizer: Literal["Euclidean", "Entropy"]="Entropy", feedback="Q", weighted=False, bidilated=True):
        super().__init__()

        self.eta = eta
        self.tau = tau
        self.gamma = gamma
        self.regularizer = regularizer
        self.feedback = feedback

        with LiteEFG.backward(is_static=True):

            self.alpha = 1.0
            if weighted:
                self.alpha = LiteEFG.const(1, 1.0)
                self.alpha.inplace(LiteEFG.aggregate(self.alpha, "sum"))
                self.alpha.inplace((self.alpha.max() + 1) * 2)

            self.ev = LiteEFG.const(1, 0.0)
            self.coef = self.tau
            self.mu = self.subtree_size.normalize(p_norm=1.0, ignore_negative=True)
            self.eta_coef = self.eta * self.coef
            
            self.u = LiteEFG.const(self.action_set_size, 1.0 / self.action_set_size)
            self.bar_u = self.u.copy()

        with LiteEFG.backward():

            if(self.feedback in ["Q", "traj-Q", "counterfactual"]):
                self._full_information()
            elif self.feedback == "Outcome":
                self._outcome_sampling()

        print("===============Graph is ready for QFR===============")
        print("eta: %f, tau: %f, gamma: %f, regularizer: %s, feedback: %s" % (self.eta, self.tau, self.gamma, self.regularizer, self.feedback))
        print("====================================================\n")
    
    def _full_information(self):
        if self.feedback == "counterfactual":
            self.m_th = 1.0
        elif self.feedback == "traj-Q":
            self.m_th = 1.0 / self.reach_prob
        else:
            self.m_th = self.opponent_reach_prob
        
        self.eta_tau = self.eta_coef / self.m_th * self.opponent_reach_prob + 1 # 1 + eta * tau / m_th * mu_{-p}(s)
        
        if self.regularizer == "Euclidean":
            reg = LiteEFG.euclidean(self.u) * self.coef
        else:
            reg = LiteEFG.negative_entropy(self.u, shifted=True) * self.coef

        bidilated_reg = reg * self.reach_prob
        gradient = LiteEFG.aggregate(self.ev, "sum") + self.utility \
                                                    + LiteEFG.aggregate(bidilated_reg, "sum", player="opponents")
        
        self.ev.inplace(LiteEFG.dot(gradient, self.u) - reg * self.opponent_reach_prob)

        gradient.inplace(gradient / self.m_th * self.eta / self.alpha)

        self._update(self.bar_u, self.bar_u, gradient)
        self._update(self.u, self.bar_u, gradient)
    
    def _outcome_sampling(self):
        #self.m_th = 1.0 / self.reach_prob
        self.eta_tau = self.eta_coef + 1 # 1 + eta * tau

        if self.regularizer == "Euclidean":
            reg = LiteEFG.euclidean(self.u) * self.coef
        else:
            reg = LiteEFG.negative_entropy(self.u, shifted=True) * self.coef

        gradient = LiteEFG.aggregate(self.ev, "sum") + self.utility \
                                                    + LiteEFG.aggregate(reg, "sum", player="opponents")
        self.ev.inplace(LiteEFG.sum(gradient) - reg)

        gradient.inplace(gradient / self.u * self.eta / self.alpha)

        self._update(self.bar_u, self.bar_u, gradient)
        self._update(self.u, self.bar_u, gradient)

    def _update(self, upd_u, ref_u, gradient):
        if self.regularizer == "Euclidean":
            upd_u.inplace((ref_u + gradient) / self.eta_tau)
            upd_u.inplace(upd_u.project(distance="L2", gamma=self.gamma, mu=self.mu))
        else:
            upd_u.inplace((ref_u.log() + gradient) / self.eta_tau)
            upd_u.inplace(upd_u - upd_u.max())
            upd_u.inplace(upd_u.exp())
            upd_u.inplace(upd_u.project(distance="KL", gamma=self.gamma, mu=self.mu))
    
    def update_graph(self, env : LiteEFG.Environment) -> None:
        env.update(self.u)
    
    def current_strategy(self) -> LiteEFG.GraphNode:
        return self.u

if __name__ == "__main__":
    import argparse
    
    parser = argparse.ArgumentParser()
    parser.add_argument("--game", type=str, default="leduc_poker(suit_isomorphism=True)")
    parser.add_argument("--iter", type=int, default=100000)
    parser.add_argument("--print_freq", type=int, default=1000)

    parser.add_argument("--eta", help="learning rate", type=float, default=0.1)
    parser.add_argument("--tau", help="regularization coefficient", type=float, default=0.001)
    parser.add_argument("--gamma", type=float, default=0.001)
    parser.add_argument("--regularizer", type=str, choices=["Euclidean", "Entropy"], default="Entropy")
    parser.add_argument("--feedback", type=str, choices=["Q", "traj-Q", "counterfactual", "Outcome"], default="Q")
    parser.add_argument("--weighted", help="weighted dilated regularizer or not", action="store_true")

    args = parser.parse_args()

    traverse_type = "Enumerate" if(args.feedback != "Outcome") else "Outcome"
    
    from utils import train
    train(graph(args.eta, args.tau, args.gamma, args.regularizer, args.feedback, args.weighted), traverse_type, "default", args.iter, args.print_freq, args.game)
    # do not need to update strategy for algorithms only need last-iterate
    # it will be faster since LiteEFG do not need to maintain the sequence-form strategy
    # update_strategy() will enumerate all infosets. When using outcome-sampling, at each iteration,
    # the algorithm will only update a trajectory of length height